#Little project for survey paper
#Import forecasts
forecasts <- readRDS("X:/SurveyData/forecasts.rds")
######################################################################################
#Library Packages
library(dplyr)
library(tidyr)
library(readr)
library(stats)
library(sparsepca)
library(glmnet)
library(foreach)
library(doParallel)
library(iterators)
library(raster)
library(zoo)
library(vars)
library(tseries)
library(data.table)  
library(parallel) 
library(tictoc) 
library(irlba)

source("X:/SurveyData/support functions/baing.R")
source("X:/SurveyData/support functions/ht_factor_functions.R")
source("X:/SurveyData/support functions/Support_function.R")
######################################################################################
#initial window size is 1259 (about 5 years)
window_initial <- 1259
#other parameters' setup
num_tickers <- 391
forecast_horizons <- c(1, 5, 20)
n_time = 1235

######################################################################################
#MSFE/RMSFE for full sample/subsamples/Combinations
#forecasts[[h_idx]][[ticker_name]] 
######################################################################################
MSFE_models <- array(NA, dim = c(5, num_tickers, 3))
for (ticker in 1:num_tickers) {
  for (h_position in 1:3) {
    error_tmp <- matrix(NA, 1235, 5)
    for (i in 1:5) {
      error_tmp[, i] <- forecasts[[h_position]][[ticker]][[i]]
      MSFE_models[i, ticker, h_position] <- mean(error_tmp[, i]^2, na.rm = TRUE)
    }
  }
}


forecasts_comb1=list()
forecasts_comb2=list()
RMSFE_comb1=list()
RMSFE_comb2=list()
MSFE_comb1=list()
MSFE_comb2=list()
hlength=length(forecast_horizons)
for (ticker in 1:num_tickers){
  forecasts_comb1[[ticker]]=matrix(NA,n_time,hlength)
  forecasts_comb2[[ticker]]=matrix(NA,n_time,hlength)
  RMSFE_comb1[[ticker]]=matrix(NA,1,hlength)
  RMSFE_comb2[[ticker]]=matrix(NA,1,hlength)
  MSFE_comb1[[ticker]]=matrix(NA,1,hlength)
  MSFE_comb2[[ticker]]=matrix(NA,1,hlength)
  for (h_position in 1:3){
    forecasts_comb1[[ticker]][,h_position]=1/2*(forecasts[[h_position]][[ticker]][[2]]+forecasts[[h_position]][[ticker]][[3]])
    forecasts_comb2[[ticker]][,h_position]=1/3*(forecasts[[h_position]][[ticker]][[2]]+forecasts[[h_position]][[ticker]][[3]]+
                                                  forecasts[[h_position]][[ticker]][[4]])
    
    RMSFE_comb1[[ticker]][,h_position]=mean((forecasts_comb1[[ticker]][, h_position])^2)/mean((forecasts[[h_position]][[ticker]][[1]])^2)
    RMSFE_comb2[[ticker]][,h_position]=mean((forecasts_comb2[[ticker]][, h_position])^2)/mean((forecasts[[h_position]][[ticker]][[1]])^2)
    MSFE_comb1[[ticker]][,h_position]=mean((forecasts_comb1[[ticker]][, h_position])^2)
    MSFE_comb2[[ticker]][,h_position]=mean((forecasts_comb2[[ticker]][, h_position])^2)
  }
}

#We make 2 subsamples
forecast_subsample1=list()
forecast_subsample2=list()
error_sub1=list()
error_sub2=list()
for (ticker in 1:num_tickers){
  MSFE_sub1=matrix(NA, 5,hlength)
  MSFE_sub2=matrix(NA, 5,hlength)
  RMSFE_sub1=matrix(NA, 4,hlength)
  RMSFE_sub2=matrix(NA, 4,hlength)
  error_sub1[[ticker]]=list()
  error_sub2[[ticker]]=list()
  for (h in forecast_horizons){
    hstep=forecast_horizons
    position_h= which(h==hstep)
    MSFE_sub1[1,position_h] = mean((forecasts[[position_h]][[ticker]][[1]][1:502])^2)
    MSFE_sub1[2,position_h] = mean((forecasts[[position_h]][[ticker]][[2]][1:502])^2)
    MSFE_sub1[3,position_h] = mean((forecasts[[position_h]][[ticker]][[3]][1:502])^2)
    MSFE_sub1[4,position_h] = mean((forecasts[[position_h]][[ticker]][[4]][1:502])^2)
    MSFE_sub1[5,position_h] = mean((forecasts[[position_h]][[ticker]][[5]][1:502])^2)
    
    RMSFE_sub1[1,position_h]=  MSFE_sub1[2,position_h]/MSFE_sub1[1,position_h]
    RMSFE_sub1[2,position_h]=  MSFE_sub1[3,position_h]/MSFE_sub1[1,position_h]
    RMSFE_sub1[3,position_h]=  MSFE_sub1[4,position_h]/MSFE_sub1[1,position_h]
    RMSFE_sub1[4,position_h]=  MSFE_sub1[5,position_h]/MSFE_sub1[1,position_h]
    
    MSFE_sub2[1,position_h] = mean((forecasts[[position_h]][[ticker]][[1]][503:1235])^2)
    MSFE_sub2[2,position_h] = mean((forecasts[[position_h]][[ticker]][[2]][503:1235])^2)
    MSFE_sub2[3,position_h] = mean((forecasts[[position_h]][[ticker]][[3]][503:1235])^2)
    MSFE_sub2[4,position_h] = mean((forecasts[[position_h]][[ticker]][[4]][503:1235])^2)
    MSFE_sub2[5,position_h] = mean((forecasts[[position_h]][[ticker]][[5]][503:1235])^2)
    
    RMSFE_sub2[1,position_h]=  MSFE_sub2[2,position_h]/MSFE_sub2[1,position_h]
    RMSFE_sub2[2,position_h]=  MSFE_sub2[3,position_h]/MSFE_sub2[1,position_h]
    RMSFE_sub2[3,position_h]=  MSFE_sub2[4,position_h]/MSFE_sub2[1,position_h]
    RMSFE_sub2[4,position_h]=  MSFE_sub2[5,position_h]/MSFE_sub2[1,position_h]
    
    error_sub1[[ticker]][[position_h]]=list(forecasts[[position_h]][[ticker]][[1]][1:502],
                                            forecasts[[position_h]][[ticker]][[2]][1:502],
                                            forecasts[[position_h]][[ticker]][[3]][1:502],
                                            forecasts[[position_h]][[ticker]][[4]][1:502],
                                            forecasts[[position_h]][[ticker]][[5]][1:502])
    error_sub2[[ticker]][[position_h]]=list(forecasts[[position_h]][[ticker]][[1]][503:1235],
                                            forecasts[[position_h]][[ticker]][[2]][503:1235],
                                            forecasts[[position_h]][[ticker]][[3]][503:1235],
                                            forecasts[[position_h]][[ticker]][[4]][503:1235],
                                            forecasts[[position_h]][[ticker]][[5]][503:1235])
  }
  forecast_subsample1[[ticker]]=list(MSFE_sub1,RMSFE_sub1)
  forecast_subsample2[[ticker]]=list(MSFE_sub2,RMSFE_sub2)
}

forecasts_comb1_sub1=list()
forecasts_comb1_sub2=list()
forecasts_comb2_sub1=list()
forecasts_comb2_sub2=list()
MSFE_comb1_sub1=list()
MSFE_comb1_sub2=list()
MSFE_comb2_sub1=list()
MSFE_comb2_sub2=list()
for (ticker in 1:391){
  forecasts_comb1_sub1[[ticker]]=matrix(NA,502,3)
  forecasts_comb1_sub2[[ticker]]=matrix(NA,733,3)
  forecasts_comb2_sub1[[ticker]]=matrix(NA,502,3)
  forecasts_comb2_sub2[[ticker]]=matrix(NA,733,3)
  MSFE_comb1_sub1[[ticker]]=matrix(NA,1,3)
  MSFE_comb1_sub2[[ticker]]=matrix(NA,1,3)
  MSFE_comb2_sub1[[ticker]]=matrix(NA,1,3)
  MSFE_comb2_sub2[[ticker]]=matrix(NA,1,3)
  for (h_position in 1:3){
    forecasts_comb1_sub1[[ticker]][,h_position]=1/2*(error_sub1[[ticker]][[h_position]][[2]]+error_sub1[[ticker]][[h_position]][[3]])
    forecasts_comb1_sub2[[ticker]][,h_position]=1/2*(error_sub2[[ticker]][[h_position]][[2]]+error_sub2[[ticker]][[h_position]][[3]])
    forecasts_comb2_sub1[[ticker]][,h_position]=1/4*(error_sub1[[ticker]][[h_position]][[2]]+error_sub1[[ticker]][[h_position]][[3]]+
                                                       error_sub1[[ticker]][[h_position]][[4]]+error_sub1[[ticker]][[h_position]][[5]])
    forecasts_comb2_sub2[[ticker]][,h_position]=1/4*(error_sub2[[ticker]][[h_position]][[2]]+error_sub2[[ticker]][[h_position]][[3]]+
                                                       error_sub2[[ticker]][[h_position]][[4]]+error_sub2[[ticker]][[h_position]][[5]])
    MSFE_comb1_sub1[[ticker]][,h_position]=mean((forecasts_comb1_sub1[[ticker]][, h_position])^2)
    MSFE_comb1_sub2[[ticker]][,h_position]=mean((forecasts_comb1_sub2[[ticker]][, h_position])^2)
    MSFE_comb2_sub1[[ticker]][,h_position]=mean((forecasts_comb2_sub1[[ticker]][, h_position])^2)
    MSFE_comb2_sub2[[ticker]][,h_position]=mean((forecasts_comb2_sub2[[ticker]][, h_position])^2)
  }
}
######################################################################################
#Average ranking and # stocks win the game
######################################################################################
bestsingle_full=list()
bestsingle_sub1=list()
bestsingle_sub2=list()
for (i in 1:391) {
  bestsingle_full[[i]]=MSFE_models[2:5,i,] #in order of PCA,SPCA,HT,LASSO, respectively
  bestsingle_sub1[[i]]=forecast_subsample1[[i]][[1]][2:5,]
  bestsingle_sub2[[i]]=forecast_subsample2[[i]][[1]][2:5,]
}
#count the number of stocks that win across four comb+single
best_single_full_n = matrix(0, nrow = 3, ncol = 4) 
best_single_sub1_n = matrix(0, nrow = 3, ncol = 4)  
best_single_sub2_n = matrix(0, nrow = 3, ncol = 4)  
model_names = c( "PCA", "SPCA", "HT","LASSO")
colnames(best_single_full_n) <- model_names
colnames(best_single_sub1_n) <- model_names
colnames(best_single_sub2_n) <- model_names

for (h in 1:3){
  counts_full = counts_sub1 = counts_sub2 = c( PCA = 0, SPCA = 0, HT = 0, LASSO = 0)
  for (i in 1:391){
    rank_full=rank(bestsingle_full[[i]][,h])
    rank_sub1=rank(bestsingle_sub1[[i]][,h])
    rank_sub2=rank(bestsingle_sub2[[i]][,h])
    best_model_full = which(rank_full == min(rank_full))
    best_model_sub1 = which(rank_sub1 == min(rank_sub1))
    best_model_sub2 = which(rank_sub2 == min(rank_sub2))
    if (1 %in% best_model_full) counts_full["PCA"] = counts_full["PCA"] + 1
    if (2 %in% best_model_full) counts_full["SPCA"] = counts_full["SPCA"] + 1
    if (3 %in% best_model_full) counts_full["HT"] = counts_full["HT"] + 1
    if (4 %in% best_model_full) counts_full["LASSO"] = counts_full["LASSO"] + 1
    
    if (1 %in% best_model_sub1) counts_sub1["PCA"] = counts_sub1["PCA"] + 1
    if (2 %in% best_model_sub1) counts_sub1["SPCA"] = counts_sub1["SPCA"] + 1
    if (3 %in% best_model_sub1) counts_sub1["HT"] = counts_sub1["HT"] + 1
    if (4 %in% best_model_sub1) counts_sub1["LASSO"] = counts_sub1["LASSO"] + 1
    
    if (1 %in% best_model_sub2) counts_sub2["PCA"] = counts_sub2["PCA"] + 1
    if (2 %in% best_model_sub2) counts_sub2["SPCA"] = counts_sub2["SPCA"] + 1
    if (3 %in% best_model_sub2) counts_sub2["HT"] = counts_sub2["HT"] + 1
    if (4 %in% best_model_sub2) counts_sub2["LASSO"] = counts_sub2["LASSO"] + 1
  }
  best_single_full_n[h, ] = counts_full
  best_single_sub1_n[h, ] = counts_sub1
  best_single_sub2_n[h, ] = counts_sub2
}

avg_ranking_matrix= matrix(NA,9,4)
for (h in 1:3){
  ranking_matrix_tmp1 <- matrix(0, nrow=391, ncol=4)
  ranking_matrix_tmp2 <- matrix(0, nrow=391, ncol=4)  
  ranking_matrix_tmp3 <- matrix(0, nrow=391, ncol=4)  
  colnames(ranking_matrix_tmp1) <- c("PCA", "SPCA", "HT","LASSO")  
  colnames(ranking_matrix_tmp2) <- c("PCA", "SPCA", "HT","LASSO")  
  colnames(ranking_matrix_tmp3) <- c("PCA", "SPCA", "HT","LASSO") 
  # Iterate through each ticker to compute rankings
  for (i in 1:391) {
    ticker_name <- names(forecasts)[i]
    msfe_matrix1 <- bestsingle_full[[i]][,h]
    msfe_matrix2 <- bestsingle_sub1[[i]][,h]
    msfe_matrix3 <- bestsingle_sub2[[i]][,h]
    # Compute rankings 
    rankings1 <- rank(msfe_matrix1)
    rankings2 <- rank(msfe_matrix2)
    rankings3 <- rank(msfe_matrix3)
    ranking_matrix_tmp1[i,] <- rankings1
    ranking_matrix_tmp2[i,] <- rankings2
    ranking_matrix_tmp3[i,] <- rankings3
    avg_ranking_matrix[h,]<- colMeans(ranking_matrix_tmp1, na.rm = TRUE)
    avg_ranking_matrix[(3+h),]<- colMeans(ranking_matrix_tmp2, na.rm = TRUE)
    avg_ranking_matrix[(6+h),]<- colMeans(ranking_matrix_tmp3, na.rm = TRUE)
  }
}
best_single_full_n
best_single_sub1_n
best_single_sub2_n
avg_ranking_matrix
######################################################################################
#Average Ranking for fullsample and subsamples 
######################################################################################
best_single_full=matrix(0,391,3)
best_single_sub1=matrix(0,391,3)
best_single_sub2=matrix(0,391,3)
avg_ranking_matrix= matrix(NA,9,4)
for (h in 1:3){
  ranking_matrix_tmp1 <- matrix(0, nrow=391, ncol=4)
  ranking_matrix_tmp2 <- matrix(0, nrow=391, ncol=4)  
  ranking_matrix_tmp3 <- matrix(0, nrow=391, ncol=4)  
  colnames(ranking_matrix_tmp1) <- c("PCA", "SPCA", "HT","LASSO")  
  colnames(ranking_matrix_tmp2) <- c("PCA", "SPCA", "HT","LASSO")  
  colnames(ranking_matrix_tmp3) <- c("PCA", "SPCA", "HT","LASSO") 
  # Iterate through each ticker to compute rankings
  for (i in 1:391) {
    ticker_name <- names(forecasts)[i]
    msfe_matrix1 <- MSFE_models[2:5,i,h]
    msfe_matrix2 <- forecast_subsample1[[i]][[1]][2:5,h]
    msfe_matrix3 <- forecast_subsample2[[i]][[1]][2:5,h]
    
    # Compute rankings 
    rankings1 <- rank(msfe_matrix1)
    rankings2 <- rank(msfe_matrix2)
    rankings3 <- rank(msfe_matrix3)
    ranking_matrix_tmp1[i,] <- rankings1
    ranking_matrix_tmp2[i,] <- rankings2
    ranking_matrix_tmp3[i,] <- rankings3
    avg_ranking_matrix[h,]<- colMeans(ranking_matrix_tmp1, na.rm = TRUE)
    avg_ranking_matrix[(3+h),]<- colMeans(ranking_matrix_tmp2, na.rm = TRUE)
    avg_ranking_matrix[(6+h),]<- colMeans(ranking_matrix_tmp3, na.rm = TRUE)
    best_model_index_full = which.min(avg_ranking_matrix[h,])
    best_model_index_sub1 = which.min(avg_ranking_matrix[(3+h),])
    best_model_index_sub2 = which.min(avg_ranking_matrix[(6+h),])
    best_single_full[i,h] = MSFE_models[(best_model_index_full+1),i,h]
    best_single_sub1[i,h] = forecast_subsample1[[i]][[1]][(best_model_index_sub1+1),h]  
    best_single_sub2[i,h] = forecast_subsample2[[i]][[1]][(best_model_index_sub2+1),h]  
  }
}
######################################################################################
#Average Ranking for combinations and best single model 
######################################################################################
comb_bestsingle_full=list()
comb_bestsingle_sub1=list()
comb_bestsingle_sub2=list()
for (i in 1:391) {
  comb_bestsingle_full[[i]]=rbind(MSFE_comb1[[i]],MSFE_comb2[[i]],best_single_full[i,])
  comb_bestsingle_sub1[[i]]=rbind(MSFE_comb1_sub1[[i]],MSFE_comb2_sub1[[i]],best_single_sub1[i,])
  comb_bestsingle_sub2[[i]]=rbind(MSFE_comb1_sub2[[i]],MSFE_comb2_sub2[[i]],best_single_sub2[i,])
}
#count the number of stocks that win across four comb+single
best_single_full_n = matrix(0, nrow = 3, ncol = 3) 
best_single_sub1_n = matrix(0, nrow = 3, ncol = 3)  
best_single_sub2_n = matrix(0, nrow = 3, ncol = 3)  
model_names = c( "Comb1", "Comb2","Single")
colnames(best_single_full_n) <- model_names
colnames(best_single_sub1_n) <- model_names
colnames(best_single_sub2_n) <- model_names

for (h in 1:3){
  counts_full = counts_sub1 = counts_sub2 = c( Comb1 = 0, Comb2 = 0, Single = 0)
  for (i in 1:391){
    rank_full=rank(comb_bestsingle_full[[i]][,h])
    rank_sub1=rank(comb_bestsingle_sub1[[i]][,h])
    rank_sub2=rank(comb_bestsingle_sub2[[i]][,h])
    best_model_full = which(rank_full == min(rank_full))
    best_model_sub1 = which(rank_sub1 == min(rank_sub1))
    best_model_sub2 = which(rank_sub2 == min(rank_sub2))
    if (1 %in% best_model_full) counts_full["Comb1"] = counts_full["Comb1"] + 1
    if (2 %in% best_model_full) counts_full["Comb2"] = counts_full["Comb2"] + 1
    if (3 %in% best_model_full) counts_full["Single"] = counts_full["Single"] + 1
    
    if (1 %in% best_model_sub1) counts_sub1["Comb1"] = counts_sub1["Comb1"] + 1
    if (2 %in% best_model_sub1) counts_sub1["Comb2"] = counts_sub1["Comb2"] + 1
    if (3 %in% best_model_sub1) counts_sub1["Single"] = counts_sub1["Single"] + 1
    
    if (1 %in% best_model_sub2) counts_sub2["Comb1"] = counts_sub2["Comb1"] + 1
    if (2 %in% best_model_sub2) counts_sub2["Comb2"] = counts_sub2["Comb2"] + 1
    if (3 %in% best_model_sub2) counts_sub2["Single"] = counts_sub2["Single"] + 1
  }
  best_single_full_n[h, ] = counts_full
  best_single_sub1_n[h, ] = counts_sub1
  best_single_sub2_n[h, ] = counts_sub2
}

avg_ranking_matrix_comb= matrix(NA,9,3)
for (h in 1:3){
  ranking_matrix_tmp1 <- matrix(0, nrow=391, ncol=3)
  ranking_matrix_tmp2 <- matrix(0, nrow=391, ncol=3)  
  ranking_matrix_tmp3 <- matrix(0, nrow=391, ncol=3)  
  colnames(ranking_matrix_tmp1) <- c("Comb1", "Comb2", "Single")  
  colnames(ranking_matrix_tmp2) <- c("Comb1", "Comb2", "Single")  
  colnames(ranking_matrix_tmp3) <- c("Comb1", "Comb2", "Single") 
  # Iterate through each ticker to compute rankings
  for (i in 1:391) {
    ticker_name <- names(forecasts)[i]
    msfe_matrix1 <- comb_bestsingle_full[[i]][,h]
    msfe_matrix2 <- comb_bestsingle_sub1[[i]][,h]
    msfe_matrix3 <- comb_bestsingle_sub2[[i]][,h]
    # Compute rankings 
    rankings1 <- rank(msfe_matrix1)
    rankings2 <- rank(msfe_matrix2)
    rankings3 <- rank(msfe_matrix3)
    ranking_matrix_tmp1[i,] <- rankings1
    ranking_matrix_tmp2[i,] <- rankings2
    ranking_matrix_tmp3[i,] <- rankings3
    avg_ranking_matrix_comb[h,]<- colMeans(ranking_matrix_tmp1, na.rm = TRUE)
    avg_ranking_matrix_comb[(3+h),]<- colMeans(ranking_matrix_tmp2, na.rm = TRUE)
    avg_ranking_matrix_comb[(6+h),]<- colMeans(ranking_matrix_tmp3, na.rm = TRUE)
  }
}
best_single_full_n
best_single_sub1_n
best_single_sub2_n
avg_ranking_matrix_comb
