###Empirical Part for Survey Paper
###Kaiwen Qiu
###Update Time: 2025-02-11
######################################################################################
#Library Packages
library(dplyr)
library(tidyr)
library(readr)
library(stats)
library(sparsepca)
library(glmnet)
library(foreach)
library(doParallel)
library(iterators)
library(raster)
library(zoo)
library(vars)
library(tseries)
library(data.table)  
library(parallel) 
library(tictoc) 
library(irlba)

source("X:/SurveyData/support functions/baing.R")
source("X:/SurveyData/support functions/ht_factor_functions.R")
source("X:/SurveyData/support functions/Support_function.R")
######################################################################################
#Data cleanning
######################################################################################
sp500_09_19 <- read.csv("X:/SurveyData/sp500_09_19.csv")
sp500_09_19$RET <- as.numeric(as.character(sp500_09_19$RET))
sp500_09_19$TICKER <- as.character(sp500_09_19$TICKER)
#add continuously compounded daily returns (log return *100, or log(pt+1/pt)*100%)
#sp500_09_19$LOG_RET <- log1p(sp500_09_19$RET)*100
# Check for empty ticker names and replace them with NA
sp500_09_19 <- sp500_09_19 %>%
  mutate(TICKER = ifelse(TICKER == "", NA, TICKER))
# Remove rows where the ticker is NA
sp500_09_19 <- filter(sp500_09_19, !is.na(TICKER))
#Transfer to list
list_by_date <- split(sp500_09_19, sp500_09_19$date)
list_by_ticker <- split(sp500_09_19, sp500_09_19$TICKER)
sp500_new <- lapply(list_by_date, function(df) as.list(df))
#create ticker list
all_tickers <- unique(unlist(lapply(sp500_new, function(x) x$TICKER)))
# Initialize the matrix with NA values
sp500_matrix <- matrix(NA, nrow = length(sp500_new), ncol = length(all_tickers))
rownames(sp500_matrix) <- names(sp500_new)  # Assign the row names as the dates from the list names
colnames(sp500_matrix) <- all_tickers  # Assign the column names as the unique tickers
#Fill the matrix
for(i in seq_along(sp500_new)) {
  # Get the current list element
  element <- sp500_new[[i]]
  
  # Extract the date from the list name
  date <- names(sp500_new)[i]
  
  # Extract the returns and tickers
  returns <- element$RET
  tickers <- as.character(element$TICKER)
  
  # Match the tickers to the column names and update the matrix
  sp500_matrix[date, tickers] <- returns
}
##Till here, we get raw return data that has all SP500 companies from 2009 t0 2019
#Next, keep the companies that has a complete return timeseries only.
complete_columns <- colSums(is.na(sp500_matrix)) <= 2769*0.05

# Subset the matrix to keep only complete columns
sp500_matrix_new <- sp500_matrix[, complete_columns]
# Linear interpolation for the rest of missing values (less than 5% for each stock)
sp500_matrix_linear <- na.approx(sp500_matrix_new)
sp500_matrix_linear2 <- na.locf(sp500_matrix_linear)
rownames(sp500_matrix_linear2) <- names(sp500_new)  # Assign the row names as the dates from the list names
######################################################################################
saveRDS(sp500_matrix_linear2, "X:/SurveyData/sp500_matrix_linear2.rds")

# Convert matrix to dataframe (optional)
df <- as.data.frame(sp500_matrix_linear2)

# Save to CSV file
write.csv(df, "X:/SurveyData/sp500_matrix_linear2.csv", row.names = FALSE)
######################################################################################
#Parameters setup
######################################################################################
#initial window size is 1259 (about 5 years)
window_initial <- 1259
#other parameters' setup
num_tickers <- ncol(sp500_matrix_linear2)
forecast_horizons <- c(1, 5, 20)
max_factor <- 20
num_factors_max =4

startpoint_f = 1299 #rownames(sp500_matrix_linear2)[1299] "2014-02-28"
endpoint_f = 2533 #rownames(sp500_matrix_linear2)[2533] "2019-01-24"

y_t = sp500_matrix_linear2
y_t_scaled = scale(y_t)
n_time = endpoint_f - startpoint_f + 1
######################################################################################
#1. PCA factor 
#output: pca_results[[h]][[t]]
# Accessing results example
# pca_results[["1_day_ahead"]][["starting_at_day_1"]]
######################################################################################
# Initialize the list structure for each forecast horizon
pca_results <- list()
for (h in forecast_horizons) {
  pca_results[[paste(h, "days_ahead", sep = "_")]] <- list()
}

# Iterate over each forecast horizon
for (h in forecast_horizons) {
  # Iterate over the time series with a rolling window
  for (start_idx in 1:n_time) {
    window_data <- y_t_scaled[(startpoint_f - window_initial - 2 * h + start_idx):(startpoint_f + start_idx -h - 1), , drop = FALSE]
    # Handle NAs and Infs
    window_data[is.na(window_data)] <- 0
    window_data[is.infinite(window_data)] <- 0
    # Determine the number of factors using Bai and Ng (2002)
    num_factors=baing(window_data,num_factors_max)
    # Perform PCA
    pca_fit = PCA_baing(window_data,num_factors)
    # Extract and store factors and loadings
    pca_results[[paste(h, "days_ahead", sep = "_")]][[paste("day", start_idx)]] =pca_fit
  }
}
######################################################################################
#2. SPCA factors
#output:spca_results[[h]][[t]]
# Accessing results example
#spca_results[["1_day_ahead"]][["starting_at_day_1"]]
######################################################################################
# Initialize the list structure for each forecast horizon
pca_results <- list()
for (h in forecast_horizons) {
  spca_results[[paste(h, "days_ahead", sep = "_")]] <- list()
}

# Iterate over each forecast horizon
for (h in forecast_horizons) {
  # Iterate over the time series with a rolling window
  for (start_idx in 1:n_time) {
    window_data <- y_t_scaled[(startpoint_f - window_initial - 2 * h + start_idx):(startpoint_f + start_idx -h - 1), , drop = FALSE]
    # Handle NAs and Infs
    window_data[is.na(window_data)] <- 0
    window_data[is.infinite(window_data)] <- 0
    # Determine the number of factors using Bai and Ng (2002)
    num_factors <- baing(window_data, num_factors_max)
    # Perform Sparse PCA
    spca_fit <- spca(X = window_data, k = num_factors, center = TRUE, scale = TRUE, verbose = FALSE)
    # Extract the factors (scores)
    factors <- spca_fit$scores
    # Save SPCA results for this target variable
    pca_results[[paste(h, "days_ahead", sep = "_")]][[paste("day", start_idx)]] =factors
  }
}
######################################################################################
#3. HT factors
#output :HT_factor_sets[[ticker]][[h]][[t]]
# Accessing results example
# HT_results_all[["starting_at_day_1"]][["1_day_ahead"]][[ticker name]]
######################################################################################
t_test_threshold <- 1.28
# Number of cores to use
num_cores <- detectCores() - 2
cl <- makeCluster(num_cores)

# Export necessary variables and functions to the cluster
clusterExport(cl, varlist = c("y_t", "forecast_horizons", "sp500_matrix_linear2", 
                              "window_initial", "startpoint_f", "num_tickers", 
                              "t_test_threshold", "max_factor", "num_factors_max", 
                              "baing", "PCA_baing",'n_time','VARselect','custom_lag','compute_ht_factors'))

# Parallelize computation over start_idx
result <- parLapply(cl, 1:n_time, function(start_idx) {
  compute_ht_factors(
    start_idx = start_idx,
    forecast_horizons = forecast_horizons,
    y_t = y_t,
    sp500_matrix_linear2 = sp500_matrix_linear2,
    window_initial = window_initial,
    startpoint_f = startpoint_f,
    num_tickers = num_tickers,
    t_test_threshold = t_test_threshold,
    max_factor = max_factor,
    num_factors_max = num_factors_max
  )
})

# Stop the cluster after computation
stopCluster(cl)
# Initialize the lists
HT_results_all <- list()
HT_selected_variables_all <- list()  #list to store selected variable counts

# Combine the results into cs_results_all and selected_variables_all
for (res in result) {
  day_position <- res$start_idx
  HT_results <- res$factors
  selected_variables_count <- res$selected_vars
  
  date_index <- day_position + startpoint_f - 1
  date <- rownames(sp500_matrix_linear2)[date_index]
  
  if (is.na(date) || date_index > length(rownames(sp500_matrix_linear2))) {
    warning(sprintf("Invalid date for day_position %d at index %d", day_position, date_index))
    HT_results_all[[as.character(day_position)]] <- NA  # Add placeholder for invalid indices
    HT_selected_variables_all[[as.character(day_position)]] <- NA
    next
  }
  
  HT_results_all[[date]] <- HT_results
  HT_selected_variables_all[[date]] <- selected_variables_count
}
######################################################################################
#4. LASSO
######################################################################################
#PARALLEL(multi-core cpu needed)
tic()
# Set up the parallel backend
num_cores <- detectCores() - 1  # Leave one core free for system processes
cl <- makeCluster(num_cores)
# Export necessary variables and functions to the cluster
clusterExport(cl, varlist = c("y_t",
                              "window_initial", "forecast_horizons",
                              "sp500_matrix_linear2", "n_time","num_tickers",
                              "startpoint_f","t_test_threshold",
                              'glmnet','cv.glmnet'))
# Parallelize the loop using parLapply over start_idx
result <- parLapply(cl, 1:n_time, function(start_idx) {
  tryCatch({
    lasso_factor_sets <- list()
    selected_variables_count <- list()
    for (h_idx in 1:length(forecast_horizons)) {
      h <- forecast_horizons[h_idx]
      lasso_factor_sets[[h_idx]] <- list()
      selected_variables_count[[h_idx]] <- list()
      for (ticker_id in 1:num_tickers) {
        gc()
        ticker_name <- colnames(sp500_matrix_linear2)[ticker_id]
        # Extract y_t_window for the current window (including all variables)
        y_tmp <- y_t[(startpoint_f - window_initial - 2*h + start_idx):(startpoint_f + start_idx - h - 1), ticker_id, drop = FALSE]
        Z_tmp <- y_t[(startpoint_f - window_initial - 2*h + start_idx):(startpoint_f + start_idx - h - 1), -ticker_id, drop = FALSE]
        cv_fit <- cv.glmnet(as.matrix(Z_tmp),y_tmp ,nfolds = 10)
        best_lambda <- cv_fit$lambda.1se
        fit1 <- glmnet(Z_tmp, y_tmp, alpha = 1, lambda = best_lambda)
        # Coefficients of the model can be accessed with
        coef1 <- coef(fit1)
        # Initialize Z_new
        Z_temp= t(as.matrix(Z_tmp))
        Z_new1 = matrix(0, nrow(Z_temp), ncol(Z_temp))
        # Fill in the coefficients for the variables selected by the LASSO
        for (j in 1:nrow(Z_temp)) {
          # Check if coef1[j] is not NA and not zero
          if (!is.na(coef1[j]) && coef1[j] != 0) {
            Z_new1[j, ] = Z_temp[j, ]
          }
        }
        Z_new1 <- Z_new1[rowSums(Z_new1 != 0) > 0, ]
        N1_tmp = nrow(Z_new1)
        # Store the result
        lasso_factor_sets[[h_idx]][[ticker_name]] <- Z_new1
        selected_variables_count[[h_idx]][[ticker_name]] <- N1_tmp
      }
    }
    return(list(start_idx = start_idx, factors = lasso_factor_sets,selected_vars = selected_variables_count))
  })
})
# Initialize the lists
lasso_results_all_1 <- list()
lasso_selected_variables_all_1 <- list()  # New list to store selected variable counts
# Combine the results into cs_results_all and selected_variables_all
for (res in result) {
  day_position <- res$start_idx
  lasso_results <- res$factors
  selected_variables_count <- res$selected_vars
  date_index <- day_position + startpoint_f - 1
  date <- rownames(sp500_matrix_linear2)[date_index]
  if (is.na(date) || date_index > length(rownames(sp500_matrix_linear2))) {
    warning(sprintf("Invalid date for day_position %d at index %d", day_position, date_index))
    lasso_results_all_1[[as.character(day_position)]] <- NA  # Add placeholder for invalid indices
    lasso_selected_variables_all_1[[as.character(day_position)]] <- NA
    next
  }
  lasso_results_all_1[[date]] <- lasso_results
  lasso_selected_variables_all_1[[date]] <- selected_variables_count
}
toc()

######################################################################################
#Forecasting Part
######################################################################################
lasso_factor_sets=lasso_results_all_1
HT_factor_sets=HT_results_all
MSFE_models = array(NA, dim = c(5, num_tickers, length(forecast_horizons)))  
RMSFE_models = array(NA, dim = c(4, num_tickers, length(forecast_horizons)))  

forecasts <- list()
# Loop over forecast horizons
for (h_idx in 1:length(forecast_horizons)) {
  h <- forecast_horizons[h_idx]
  cat("Processing forecast horizon:", h, "\n")
  forecasts[[h_idx]]=list()
  # Loop over each ticker
  for (ticker_id in 1:num_tickers) {
    gc()
    ticker_name <- colnames(sp500_matrix_linear2)[ticker_id]
    cat("  Processing ticker:", ticker_name, "\n")
    
    # Initialize storage for forecasts and errors
    y_t_Plus_1_benchmark <- numeric(n_time)
    y_t_Plus_1_pca <- numeric(n_time)
    y_t_Plus_1_spca <- numeric(n_time)
    y_t_Plus_1_ht <- numeric(n_time)
    y_t_Plus_1_lasso <- numeric(n_time)
    
    forecast_error_benchmark <- numeric(n_time)
    forecast_error_pca <- numeric(n_time)
    forecast_error_spca <- numeric(n_time)
    forecast_error_ht <- numeric(n_time)
    forecast_error_lasso <- numeric(n_time)
    
    # Loop over each time point using recursive window
    for (t in 1:n_time) {
      current_time <- startpoint_f + t - 1
      
      # Use data from the beginning up to current_time - h
      window_end <- current_time - h 
      
      y <- y_t[(startpoint_f - window_initial - 2*h + t):window_end, ticker_id , drop = FALSE]
      
      # Determine AR lags using SIC
      sic_info <- VARselect(y, lag.max = 4, type = "const")
      optimal_lags <- sic_info$selection['SC(n)']
      y_tmp= y[2:(nrow(y)-h+1)]
      if (optimal_lags > 0) {
        ylag <- NULL
        for (lag in 1:optimal_lags) {
          lagged_data <- custom_lag(y_tmp, lag)
          ylag <- cbind(ylag, lagged_data[1:length(y_tmp)])
        }
      } else {
        ylag <- NULL
      }
      ylag=ylag[(optimal_lags+1):length(y_tmp),,drop = FALSE]
      ## Get factors for current time and t-h for model fitting
      # PCA factors
      exo_pca_tmp <-  pca_results[[h_idx]][[t]][ , drop = FALSE] 
      exo_pca = exo_pca_tmp[(optimal_lags+1):(nrow(exo_pca_tmp)-h),]
      # SPCA factors
      exo_spca_tmp <- spca_results[[h_idx]][[t]][ , drop = FALSE] 
      exo_spca = exo_spca_tmp[(optimal_lags+1):(nrow(exo_spca_tmp)-h),]
      # HT factors with null handling and dynamic length
      if (is.null(HT_factor_sets[[t]][[1]][[ticker_id]]) || 
          length(HT_factor_sets[[t]][[1]][[ticker_id]]) == 0 ||  # Check if empty
          all(is.na(HT_factor_sets[[t]][[1]][[ticker_id]])) ||  # Check if all NA
          is.null(dim(HT_factor_sets[[t]][[1]][[ticker_id]])) ||  # Check if it has no valid dimensions
          nrow(HT_factor_sets[[t]][[1]][[ticker_id]]) == 0 ||  # Check if empty matrix
          ncol(HT_factor_sets[[t]][[1]][[ticker_id]]) == 0) {  # Check if empty matrix
        exo_ht <- NULL
      } else {
        exo_ht_tmp <- HT_factor_sets[[t]][[1]][[ticker_id]]
        exo_ht <- exo_ht_tmp[(optimal_lags+1):(nrow(exo_ht_tmp)-h), , drop=FALSE]
      }
      #LASSO factors with null handling
      if (is.null(lasso_factor_sets[[t]][[1]][[ticker_id]]) || 
          length(lasso_factor_sets[[t]][[1]][[ticker_id]]) == 0 ||  # Check if empty
          all(is.na(lasso_factor_sets[[t]][[1]][[ticker_id]])) ||  # Check if all NA
          is.null(dim(lasso_factor_sets[[t]][[1]][[ticker_id]])) ||  # Check if it has no valid dimensions
          nrow(lasso_factor_sets[[t]][[1]][[ticker_id]]) == 0 ||  # Check if empty matrix
          ncol(lasso_factor_sets[[t]][[1]][[ticker_id]]) == 0) {  # Check if empty matrix
        exo_lasso <- NULL
      } else {
        exo_lasso_tmp <- t(lasso_factor_sets[[t]][[1]][[ticker_id]])
        exo_lasso <- exo_lasso_tmp[(optimal_lags+1):(nrow(exo_lasso_tmp)-h), , drop=FALSE]
      }
      min_rows = min(nrow(ylag), nrow(exo_pca), nrow(exo_spca), ifelse(is.null(exo_ht), Inf, nrow(exo_ht)),
                     ifelse(is.null(exo_lasso), Inf, nrow(exo_lasso)))
      # Adjust rows dynamically by removing the oldest rows
      rows_to_remove <- nrow(ylag) - min_rows  # Difference in lengths
      
      if (rows_to_remove > 0) {
        ylag = ylag[-(1:rows_to_remove), , drop = FALSE]
        exo_pca = exo_pca[-(1:rows_to_remove), , drop = FALSE]
        exo_spca = exo_spca[-(1:rows_to_remove), , drop = FALSE]
      }
      
      # Prepare the design matrices with updated lengths
      if (optimal_lags > 0) {
        X_benchmark <- cbind(1, ylag)
        X_pca <- cbind(1, ylag, exo_pca)
        X_spca <- cbind(1, ylag, exo_spca)
        if (!is.null(exo_ht)) {
          X_ht <- cbind(1, ylag, exo_ht)
        } else {
          X_ht <- cbind(1, ylag)  # Revert to benchmark if HT is null
        }
        if (!is.null(exo_lasso)){
          X_lasso = cbind(1,ylag,exo_lasso)
        } else {
          X_lasso <- cbind(1, ylag)  # Revert to benchmark if lasso is null
        }
      } else {
        X_benchmark <- cbind(1)
        X_pca <- cbind(1, exo_pca)
        X_spca <- cbind(1, exo_spca)
        if (!is.null(exo_ht)) {
          X_ht <- cbind(1, exo_ht)
        } else {
          X_ht <- cbind(1)
        }
        if (!is.null(exo_lasso)) {
          X_lasso <- cbind(1, exo_lasso)
        } else {
          X_lasso <- cbind(1)
        }
      }
      
      # Adjust dependent variable (y) in the same way
      y <- y[(h+optimal_lags+1):length(y)]      
      if (rows_to_remove > 0) {
        y = y[-(1:rows_to_remove)]
      }
      # Ensure no missing values
      X_benchmark <- X_benchmark[complete.cases(X_benchmark), ]
      X_pca <- X_pca[complete.cases(X_pca), ]
      X_spca <- X_spca[complete.cases(X_spca), ]
      X_ht <- X_ht[complete.cases(X_ht), ]
      X_lasso <- X_lasso[complete.cases(X_lasso), ]
      
      # Fit models
      tmpbeta_benchmark <- solve(t(X_benchmark) %*% X_benchmark) %*% t(X_benchmark) %*% y
      tmpbeta_pca <- solve(t(X_pca) %*% X_pca) %*% t(X_pca) %*% y
      tmpbeta_spca <- solve(t(X_spca) %*% X_spca) %*% t(X_spca) %*% y
      tmpbeta_ht <- solve(t(X_ht) %*% X_ht) %*% t(X_ht) %*% y
      tmpbeta_lasso <- solve(t(X_lasso) %*% X_lasso) %*% t(X_lasso) %*% y
      y_te=NULL
      for (i in 1:optimal_lags){
        y_te=c(y_te,y[length(y)-i+1])
      }
      # Forecast with observation at t-h
      if (optimal_lags > 0) {
        y_t_Plus_1_benchmark[t] <- c(1, y_te) %*% tmpbeta_benchmark
        y_t_Plus_1_pca[t] <- c(1, y_te, exo_pca_tmp[nrow(exo_pca_tmp), ]) %*% tmpbeta_pca
        y_t_Plus_1_spca[t] <- c(1, y_te, exo_spca_tmp[nrow(exo_spca_tmp), ]) %*% tmpbeta_spca
        if (!is.null(exo_ht)) {
          y_t_Plus_1_ht[t] <- c(1, y_te, exo_ht_tmp[nrow(exo_ht_tmp),]) %*% tmpbeta_ht
        } else {
          y_t_Plus_1_ht[t] <- c(1,y_te) %*% tmpbeta_benchmark  # Revert to benchmark if HT is null
        }
        if (!is.null(exo_lasso)) {
          y_t_Plus_1_lasso[t] <- c(1, y_te, exo_lasso_tmp[nrow(exo_lasso_tmp),]) %*% tmpbeta_lasso
        } else {
          y_t_Plus_1_lasso[t] <- c(1,y_te) %*% tmpbeta_benchmark  # Revert to benchmark if HT is null
        }
      } else {
        y_t_Plus_1_benchmark[t] <- c(1) %*% tmpbeta_benchmark
        y_t_Plus_1_pca[t] <- c(1, exo_pca_tmp[nrow(exo_pca_tmp), ]) %*% tmpbeta_pca
        y_t_Plus_1_spca[t] <- c(1, exo_spca[nrow(exo_spca_tmp), ]) %*% tmpbeta_spca
        if (!is.null(exo_ht)) {
          y_t_Plus_1_ht[t] <- c(1, exo_ht_tmp[nrow(exo_ht_tmp),]) %*% tmpbeta_ht
        } else {
          y_t_Plus_1_ht[t] <- c(1) %*% tmpbeta_benchmark  # Revert to benchmark if HT is null
        }
        if (!is.null(exo_lasso)) {
          y_t_Plus_1_lasso[t] <- c(1, exo_lasso_tmp[nrow(exo_lasso_tmp),]) %*% tmpbeta_lasso
        } else {
          y_t_Plus_1_lasso[t] <- c(1) %*% tmpbeta_benchmark  # Revert to benchmark if HT is null
        }
      }
      
      # Compute forecast errors
      actual_value <- y_t[current_time, ticker_id]
      forecast_error_benchmark[t] <- y_t_Plus_1_benchmark[t] - actual_value
      forecast_error_pca[t] <- y_t_Plus_1_pca[t] - actual_value
      forecast_error_spca[t] <- y_t_Plus_1_spca[t] - actual_value
      forecast_error_ht[t] <- y_t_Plus_1_ht[t] - actual_value
      forecast_error_lasso[t] <- y_t_Plus_1_lasso[t] - actual_value
      
    }
    # Compute MSFE for each model
    MSFE_benchmark <- mean(forecast_error_benchmark^2, na.rm = TRUE)
    MSFE_pca <- mean(forecast_error_pca^2, na.rm = TRUE)
    MSFE_spca <- mean(forecast_error_spca^2, na.rm = TRUE)
    MSFE_ht <- mean(forecast_error_ht^2, na.rm = TRUE)
    MSFE_lasso <- mean(forecast_error_lasso^2, na.rm = TRUE)
    
    # Store MSFE and RMSFE for each model
    MSFE_models[1, ticker_id, h_idx] <- MSFE_benchmark
    MSFE_models[2, ticker_id, h_idx] <- MSFE_pca
    MSFE_models[3, ticker_id, h_idx] <- MSFE_spca
    MSFE_models[4, ticker_id, h_idx] <- MSFE_ht
    MSFE_models[5, ticker_id, h_idx] <- MSFE_lasso
    
    RMSFE_models[1,ticker_id,h_idx]=  MSFE_pca/MSFE_benchmark
    RMSFE_models[2,ticker_id,h_idx]=  MSFE_spca/MSFE_benchmark
    RMSFE_models[3,ticker_id,h_idx]=  MSFE_ht/MSFE_benchmark
    RMSFE_models[4,ticker_id,h_idx]=  MSFE_lasso/MSFE_benchmark
    
    # Store forecast errors 
    forecasts[[h_idx]][[ticker_name]] <- list(
      benchmark = forecast_error_benchmark,
      pca = forecast_error_pca,
      spca = forecast_error_spca,
      ht = forecast_error_ht,
      lasso = forecast_error_lasso
    )
  }
}
######################################################################################
saveRDS(forecasts, "X:/SurveyData/forecasts.rds")
######################################################################################
#ENDDING!
######################################################################################


