# Function to compute HT factors for a single start index
compute_ht_factors <- function(start_idx, forecast_horizons, y_t, sp500_matrix_linear2, window_initial, 
                               startpoint_f, num_tickers, t_test_threshold, max_factor, num_factors_max) {
  HT_factor_sets <- list()
  selected_variables_count <- list()
  
  for (h_idx in seq_along(forecast_horizons)) {
    h <- forecast_horizons[h_idx]
    HT_factor_sets[[h_idx]] <- list()
    selected_variables_count[[h_idx]] <- list()
    
    for (ticker_id in seq_len(num_tickers)) {
      ticker_name <- colnames(sp500_matrix_linear2)[ticker_id]
      
      # Extract the dependent variable
      y <- y_t[
        (startpoint_f - window_initial - 2 * h + start_idx):(startpoint_f + start_idx - h - 1), 
        ticker_id, 
        drop = FALSE
      ]
      
      # Extract independent variables excluding the current target
      X <- y_t[(startpoint_f - window_initial - 2*h + start_idx):(startpoint_f + start_idx - h - 1), 
               -ticker_id, 
               drop = FALSE]
      
      # Calculate t-values for independent variables
      t_values <- numeric(ncol(X))
      names(t_values) <- colnames(X)
      
      for (col in colnames(X)) {
        x_col <- as.matrix(X[, col, drop = FALSE])
        x_design <- cbind(1, x_col)
        y <- as.matrix(y)
        
        # Determine AR lags using SIC
        sic_info <- VARselect(x_design, lag.max = 4, type = "const")
        optimal_lags <- sic_info$selection['SC(n)']
        
        if (optimal_lags > 0) {
          ylag <- NULL
          for (lag in 1:optimal_lags) {
            lagged_data <- custom_lag(y, lag)
            ylag <- cbind(ylag, lagged_data[1:length(y)])
          }
        } else {
          ylag <- NULL
        }
        y2 = cbind(x_design,ylag)
        fit = lm(y~y2)
        fit2=summary(fit)
        t_value=fit2$coefficients[2,3] #to get t value for this variable as I put x_design, so number of lags doesn't matter
        t_values[col] <- t_value
      }
      
      # Select variables based on the t-test threshold
      selected_vars <- names(t_values)[abs(t_values) > t_test_threshold]
      
      # Check if the number of selected variables exceeds the threshold
      if (length(selected_vars) > max_factor) {
        pca_data <- scale(X[, selected_vars, drop = FALSE])
        num_factors <- baing(pca_data, num_factors_max)
        pca_fit <- PCA_baing(pca_data, num_factors)
        HT_factor_sets[[h_idx]][[ticker_id]] <- pca_fit
      } else {
        HT_factor_sets[[h_idx]][[ticker_id]] <- list("factors" = NA, "loadings" = NA)
      }
      
      selected_variables_count[[h_idx]][[ticker_id]] <- length(selected_vars)
    }
  }
  
  return(list(start_idx = start_idx,factors = HT_factor_sets, selected_vars = selected_variables_count))
}
